

class PassiveWait
{
  public:
    PassiveWait () : m_setpoint(0), m_wait(0) {}
  
    void setwait ( unsigned long microseconds = 0 )
    {
      m_setpoint = micros();
      m_wait = microseconds;
    }
    
    void wait ()
    {
      if( m_wait == 0 )
        return;
      unsigned long waitpoint = m_setpoint + m_wait;
      m_wait = 0;
      if( waitpoint < m_setpoint ) // overflow expected
      {
        while( micros() > m_setpoint ){} // wait for overflow
        while( micros() < waitpoint ){}
      }
      else // no overflow expected
      {
        unsigned long t = micros();
        if( t > m_setpoint ) // no overflow
          if( t < waitpoint )
            while( micros() < waitpoint ){}
        // else, unexpected overflow
        // so time _has_ expired
      }
    }
    
  private:
    unsigned long m_setpoint;
    unsigned long m_wait;
};


/*
  LCD module wiring
  8 bit data on shift register, MSb first
  Instr/Data on shift register input
  so, shift out D7...D0, then set I/D but don't clock
    data         PORTD bit 7 ( pin 7 )
    shift clock  PORTD bit 4 ( pin 4 )
    lcd enable   PORTD bit 2 ( pin 2 )
  lcd read/write tied low (write only)
*/
class LCD
{
  public:
    
    void init ();
    
    void clear ();
    
    void customchar ( unsigned char code, char bitmap[8] );
      // code is [0,7]
      // bitmap is row major ( least 5 bits of each char )
    
    void customchar_90 ( unsigned char code, char bitmap[5] );
      // bitmap is column major
    
    void cursorpos ( int row = 0, int col = 0 );
    
    void print1 ( char c );
    
    void print ( char * s );
    
  private:
    void rawwrite ( char data, char instr = 0, unsigned int wait = 50 );
      // instr = 0: write data. else write instruction
      // wait: microseconds to allow for execution
  
    PassiveWait waiter;
};

void LCD::init ()
{
  // should wait about 40ms after lcd powers on
  // but arduino takes a long while to start, so don't bother
  PORTD &= 0b01101111; // data/clock low
  PORTD |= 0b00000100; // enable high
  DDRD  |= 0b10010100;
  
  char mode = 0b00111000; // 8 bit, 1/16 duty, 5x7 dots
  rawwrite( mode, 1, 5000 );
  rawwrite( mode, 1, 200 );
  rawwrite( mode, 1, 200 ); // three times just to set 8 bit
  rawwrite( mode, 1, 200 ); // actually set everything
  rawwrite( 0b00010000, 1 ); // cursor move, no display shift
  rawwrite( 0b00001100, 1 ); // display on, cursor off, blink off
  rawwrite( 0b00000110, 1 ); // cursor move right, no shift
  clear();
}

void LCD::rawwrite ( char data, char instr, unsigned int wait )
{
  waiter.wait();
  PORTD |= 0b00000100; // enable high

  for( int i = 0; i < 8; i++ )
  {
    // set shift input data, set clock 0
    PORTD = ( PORTD & 0b01101111 ) | ( data & 0b10000000 );
    PORTD |= 0b00010000; // clock 0->1
    data = data << 1;
  }

  if( instr == 0 )
    PORTD |= 0b10000000; // RS high (data)
  else
    PORTD &= 0b01111111; // RS low (instruction)
  
  PORTD &= 0b11111011; // enable low
  waiter.setwait( wait );
}

void LCD::clear ()
{
  rawwrite( 0b00000001, 1, 2000 );
}

void LCD::customchar ( unsigned char code, char bitmap[8] )
{
  rawwrite( 0b01000000 | ( ( code & 0b111 ) << 3 ), 1 );
  for( int row = 0; row < 8; row++ )    
    rawwrite( bitmap[ row ] & 0b11111 );
}

void LCD::customchar_90 ( unsigned char code, char bitmap[5] )
{
  rawwrite( 0b01000000 | ( ( code & 0b111 ) << 3 ), 1 );
  for( int row = 0; row < 8; row++ )
  {
    char rowbmp = 0;
    for( int col = 0; col < 5; col++ )
      rowbmp |= ( ( bitmap[col] >> ( 7 - row ) ) & 1 ) << ( 4 - col );
    rawwrite( rowbmp );
  }
}

void LCD::cursorpos ( int row, int col )
{
  unsigned char addr = col % 20;
  // row base addresses
  // 0 : 0x0, 1 : 0x40, 2 : 0x14, 3 : 0x54
  if( row & 1 )
    addr += 0x40;
  if( row & 2 )
    addr += 0x14;
  rawwrite( 0b10000000 | addr, 1 );
}

void LCD::print1 ( char c )
{
  rawwrite( c );
}

void LCD::print ( char * s )
{
  while( *s != 0 )
    rawwrite( *s++ );
}


#define P_SOP      (unsigned char)254
#define P_LCD_A    (unsigned char)100
#define P_LCD_A_CG (unsigned char)101
/*
protocol, char values

P_SOP : start of packet

P_LCD_A : write 4x20 lcd
  row
  col
  string...

P_LCD_A_CG : write 4x20 lcd custom char
  index / char code
  bitmap[5], column major
  
*/

unsigned char recv ()
{
  int c = -1;
  while( c < 0 )
    c = Serial.read();
  return c;
}

void setup ()
{
  Serial.begin( 115200 );
  
  LCD lcd;
  lcd.init();
  
  lcd.cursorpos( 0, 0 );
  lcd.print( "01234567890123456789" );
  lcd.cursorpos( 1, 10 );
  for( char x = 0; x < 8; x++ )
    lcd.print1( x );
  
  unsigned char c = 0; // immediate receive value
  for(;;)
  {
    while( c != P_SOP )
      c = recv();
    while( c == P_SOP )
      c = recv();
    
    int row, col;
    unsigned char x;
    char buf[ 8 ];
    switch( c )
    {
      case P_LCD_A:
        c = recv(); if( c == P_SOP ) break;
        row = c;
        c = recv(); if( c == P_SOP ) break;
        col = c;
        lcd.cursorpos( row, col );
        for(;;)
        {
          c = recv(); if( c == P_SOP ) break;
          lcd.print1( c );
        }
        break;
      
      case P_LCD_A_CG:
        c = recv(); if( c == P_SOP ) break;
        x = c;
        for( int i = 0; i < 5; i++ )
          buf[ i ] = recv(); // no choice here but to ignore P_SOP
        lcd.customchar_90( x, buf );
        break;
    }
  }
}

void loop () {}

