#!/usr/bin/env python

import time, serial
import math, random
import subprocess, sys, os

##############################################################################

class LCD_State:
	
	def __init__ ( self, rows = 4, columns = 20, customchars = 8 ):
		self.text = [ ' ' * columns for i in xrange(rows) ]
		self.chars = [ [ '*' * 5 for y in xrange(8) ] for c in xrange(customchars) ]
	
	def updatechar ( self, code, charbmp ):
		"""return true iff the character is actually new/different"""
		if self.chars[ code ] == charbmp:
			return False
		self.chars[ code ] = charbmp
		return True
	
	def updatetext ( self, row = 0, col = 0, s = '', mingap = 4 ):
		"""return list of tuples of updates (row,col,string)"""
		width = len( self.text[0] )
		if row < 0 or row >= len( self.text ) or col < 0 or col >= width:
			return []
		s = s[ : width - col ]
		if not s:
			return []
		
		oldrow = self.text[ row ]
		newrow = oldrow[ : col ] + s + oldrow[ col + len(s) : ]
		
		diffs = []
		for col in xrange( width ):
			if newrow[col] != oldrow[col]:
				diffs.append( col )
		if not diffs:
			return []
		# find spans of text to update
		# don't break up spans unless the gap is larger than a new header would
		#  be anyway
		spans = [ ( diffs[0], diffs[0] ) ]
		for col in diffs[ 1 : ]:
			if col > spans[-1][1] + mingap:
				spans.append( ( col, col ) )
			else:
				spans[-1] = ( spans[-1][0], col )
		
		self.text[ row ] = newrow
		return [ ( row, span[0], newrow[ span[0] : span[1] + 1 ] ) for span in spans ]


class PanelLink:
	
	def __init__ ( self, portfile = '/dev/ttyUSB0' ):
		self.serport = serial.Serial( portfile, baudrate = 115200 )
		time.sleep( 2.5 )
		# opening the serial port triggers a uC reset
		# haven't figured out how to configure the port to prevent this
		# so just give the thing some time to finish the reset
		self.lcdstate = LCD_State( 4, 20, 8 )
	
	def sync ( self ):
		for row in xrange( len( self.lcdstate.text ) ):
			self.LCD_putstring( row, 0, self.lcdstate.text[ row ], True )
		for code in xrange( len( self.lcdstate.chars ) ):
			self.LCD_customchar( code, self.lcdstate.chars[ code ], True )
	
	def LCD_putstring ( self, row = 0, col = 0, s = '', force = False ):
		updates = self.lcdstate.updatetext( row, col, s, 4 )
		if not force:
			for u in updates:
				self.serport.write( chr(254) + chr(100) )
				self.serport.write( ( chr(u[0]) + chr(u[1]) + u[2] ).replace( chr(254), ' ' ) )
		else:
			self.serport.write( chr(254) + chr(100) )
			self.serport.write( ( chr(row) + chr(col) + s ).replace( chr(254), ' ' ) )
	
	def LCD_putstring_bulk( self, strings, force = False ):
		"""strings is dict {(row,col):string}"""
		if not strings:
			return
		for pos, s in strings.iteritems():
			self.LCD_putstring( pos[0], pos[1], s, force )
	
	# code is [0,7]
	# bitmap is 8 strings length 5, ' ' is off pixel all other chars on
	def LCD_customchar ( self, code = 0, bitmap = [], force = False ):
		if not self.lcdstate.updatechar( code, bitmap ) and not force:
			return
		bitmap90 = ''
		for col in xrange( 5 ):
			colbmp = 0
			for row in xrange( 8 ):
				if bitmap[ row ][ col ] != ' ':
					colbmp += 2 ** ( 7 - row )
			bitmap90 += chr(colbmp)
		self.serport.write( chr(254) + chr(101) )
		self.serport.write( chr(code) + bitmap90 )
	
	def LCD_customchar_bulk ( self, chars ):
		"""chars is dict {code:bitmap}, as in LCD_customchar"""
		if not chars:
			return
		for c, b in chars.iteritems():
			self.LCD_customchar( c, b )


##############################################################################

class SystemStats:
	
	def __init__ ( self ):
		self.cpulastvals = self.CPU_currentvals()
		self.disklastvals = self.disk_currentvals()
		self.netlastvals = self.net_currentvals()
	
	def CPU_load ( self ):
		"""update and return cpu load [0,1]"""
		load = 0.0
		nowvals = self.CPU_currentvals()
		if sum( self.cpulastvals ) > 0 and nowvals[1] != self.cpulastvals[1]:
			load = ( nowvals[0] - self.cpulastvals[0] ) / float( nowvals[1] - self.cpulastvals[1] )
		self.cpulastvals = nowvals
		return load
	
	def CPU_currentvals ( self ):
		"""return current values for all cpus [ busytime, totaltime ]"""
		values = [ 0, 0 ]
		allvals = []
		for statline in open( '/proc/stat' ):
			if statline.startswith( 'cpu ' ):
				for stringval in statline.split()[ 1: ]:
					allvals.append( long( stringval ) )
				break
		if not allvals: return values
		values[ 1 ] = sum( allvals )
		values[ 0 ] = values[ 1 ] - sum( allvals[ 3 : 5 ] ) # omit idle and iowait
		return values
	
	def CPU_temp ( self ):
		coretemps = []
		sensors = subprocess.Popen( ['sensors'], stdout = subprocess.PIPE )
		sensors.wait()
		for s in sensors.stdout:
			if s.startswith( 'Core ' ):
				s = s[ s.find('+') + 1 : ]
				s = s[ : s.find('\xc2\xb0') ]
				coretemps.append( float( s ) )
		if not coretemps:
			return 0
		tempc = sum( coretemps ) / float( len(coretemps) )
		return int( tempc * 9./5. + 32 )
	
	def mem_load ( self ):
		"""return (ram,swap), both are in [0,100]"""
		meminfo = {}
		for procline in open( '/proc/meminfo' ):
			procparts = procline.split()
			meminfo[ procparts[0].rstrip(':') ] = long( procparts[1] )
		ram = 100 * ( meminfo['MemTotal'] - meminfo['MemFree'] - meminfo['Cached'] - meminfo['Buffers'] ) / meminfo['MemTotal']
		swap = 0
		if meminfo['SwapTotal'] > 0:
			swap = 100 * ( meminfo['SwapTotal'] - meminfo['SwapFree'] ) / meminfo['SwapTotal']
		return ( ram, swap )
	
	def disk_load ( self ):
		""" return dict, disk device : activity (boolean)"""
		loads = {}
		nowvals = self.disk_currentvals()
		for dev in nowvals.keys():
			if dev not in self.disklastvals:
				loads[ dev ] = True
			else:
				loads[ dev ] = ( nowvals[dev] != self.disklastvals[dev] )
		self.disklastvals = nowvals
		return loads
	
	def disk_currentvals ( self ):
		"""return dict, disk device : sector read/write total"""
		disks = {}
		for procline in open( '/proc/diskstats' ):
			procparts = procline.split()
			if len(procparts[2]) != 3:
				continue
			if not procparts[2].startswith( 'sd' ):
				continue
			# device is of form 'sdX'
			disks[ procparts[2] ] = long( procparts[5] ) + long( procparts[9] )
		return disks
	
	def net_load ( self ):
		"""return tuple bytes/sec (recv,send)"""
		r = 0
		s = 0
		nowvals = self.net_currentvals()
		t = nowvals[0] - self.netlastvals[0]
		if t > 0:
			r = ( nowvals[1] - self.netlastvals[1] ) / t
			s = ( nowvals[2] - self.netlastvals[2] ) / t
		self.netlastvals = nowvals
		return ( r, s )
	
	def net_currentvals ( self ):
		"""return tuple (time, recv bytes, sent bytes)"""
		dev = 'eth0'
		r = 0
		s = 0
		for procline in open( '/proc/net/dev' ):
			procline = procline.lstrip()
			if procline.startswith( dev ):
				procparts = procline.lstrip( dev + ':' ).split()
				r = long( procparts[0] )
				s = long( procparts[8] )
				break
		return ( time.time(), r, s )
	
	def uptime ( self ):
		"""return system uptime in seconds"""
		secs = open( '/proc/uptime' ).readline().split()[0]
		return long( secs.split('.')[0] )

	def fs_free ( self, path = '/' ):
		"""return bytes free at path"""
		stat = os.statvfs( path )
		return stat.f_bsize * stat.f_bfree


##############################################################################

class Animation:
	def __init__ ( self ):
		pass
	
	def step ( self, x ):
		"""update to next frame with current dependent variable"""
	
	def updatepanel ( self, panel, init = False ):
		"""push updated chars and strings to panel"""
		panel.LCD_customchar_bulk( self.charupdate( init ) )
		panel.LCD_putstring_bulk( self.stringupdate( init ) )
	
	def charupdate ( self, init = False ):
		"""return dict {code:bitmap}"""
	
	def stringupdate ( self, init = False ):
		"""return dict {(row,col):string}"""


class LoadLevels( Animation ):
	
	def __init__ ( self ):
		self.rows = 2
		self.levels = [ 0 ] * 4
		self.chars = ' \x00\x01\x02\x03\x04\x05\x06\x07'
	
	def step ( self, x ):
		"""x is [0,1]"""
		ll = self.levels
		targetsum = int( 4 * 8 * self.rows * x )
		cursum = sum( ll )
		inc = max( -1, min( 1, targetsum - cursum ) )
		if inc == 0:
			if max( ll ) - min( ll ) < 3:
				return
			maxidx = 0
			minidx = 0
			for idx in xrange( 1, len(ll) ):
				maxidx = idx if ll[ idx ] > ll[ maxidx ] else maxidx
				minidx = idx if ll[ idx ] < ll[ minidx ] else minidx
			ll[ maxidx ] -= 1
			ll[ minidx ] += 1
		else:
			col = random.randint( 0, len(ll) - 1 )
			if inc > 0:
				while ll[ col ] >= self.rows * 8:
					col = ( col + 1 ) % len(ll)
			if inc < 0:
				while ll[ col ] <= 0:
					col = ( col + 1 ) % len(ll)
			ll[ col ] += inc
	
	def charupdate ( self, init = False ):
		"""return dict {code:bitmap}"""
		if not init:
			return {}
		chars = {}
		for x in xrange( 8 ):
			bmp = [ '     ' ] * ( 7 - x ) + [ ' *** ' ] * ( x + 1 )
			chars[ x ] = bmp
		return chars
	
	def stringupdate ( self, init = False ):
		"""return dict {(row,col):string}"""
		rowstr = [ '' ] * self.rows
		for l in self.levels:
			l = max( 1, l )
			for r in xrange( self.rows ):
				cidx = l - ( self.rows - 1 - r ) * 8
				cidx = max( 0, min( 8, cidx ) )
				rowstr[ r ] += self.chars[ cidx ]
		strd = {}
		for r in xrange( self.rows ):
			strd[ (r,16) ] = rowstr[ r ]
		return strd


class LoadBlob( Animation ):
	
	def __init__ ( self ):
		Animation.__init__( self )
		
		self.img = [ [ 0 for y in xrange(16) ] for x in xrange(20) ]
		for x in xrange( 20 ):
			self.img[ x ][ 0 ] = 1
			self.img[ x ][ 15 ] = 1
		for y in xrange( 16 ):
			self.img[ 0 ][ y ] = 1
			self.img[ 19 ][ y ] = 1
		self.dotsoff = []
		for x in xrange( 1, 19 ):
			for y in xrange( 1, 15 ):
				self.dotsoff.append( ( x, y ) )
		self.dotson = []
	
	def step ( self, x ):
		"""x is [0,1]"""
		oncount = int( x * ( 18 * 14 ) )
		for n in xrange( 1 if abs( oncount - len(self.dotson) ) < 10 else 5 ):
			if oncount > len( self.dotson ):
				i = self.finddot( self.dotsoff, 1 )
				idot = self.dotsoff[ i ]
				self.img[ idot[0] ][ idot[1] ] = 1
				self.dotson.append( idot )
				del self.dotsoff[ i ]
			if oncount < len( self.dotson ):
				i = self.finddot( self.dotson, 0 )
				idot = self.dotson[ i ]
				self.img[ idot[0] ][ idot[1] ] = 0
				self.dotsoff.append( idot )
				del self.dotson[ i ]
			if oncount == len( self.dotson ):
				break
		
		#for i in xrange( max( 1, int( 50 * load ) ) ):
			#x = random.randint( 1, 18 )
			#y = random.randint( 1, 14 )
			#if random.random() < ( 0.97 - load ** 1.5 ):
				#self.img[x][y] = 0
			#else:
				#self.img[x][y] = 1 - self.img[x][y]
	
	def charupdate ( self, init = False ):
		chars = {}
		for chary in xrange( 2 ):
			for charx in xrange( 4 ):
				bmp = []
				for y in xrange( chary * 8, ( chary + 1 ) * 8 ):
					bmp += [ '' ]
					for x in xrange( charx * 5, ( charx + 1 ) * 5 ):
						bmp[ -1 ] += ' ' if self.img[x][y] == 0 else '*'
				chars[ chary * 4 + charx ] = bmp
		return chars
	
	def stringupdate ( self, init = False ):
		if init:
			return {
				(0,16) : '\x00\x01\x02\x03',
				(1,16) : '\x04\x05\x06\x07' }
		return {}
	
	def finddot ( self, dotlist = [], bias = 0 ):
		windowsize = 10
		
		if not dotlist:
			return None
		if bias == 0:
			return random.randint( 0, len(dotlist) - 1 )
		
		window = range( len(dotlist) )
		if windowsize < len(window):
			window = random.sample( window, windowsize )
		else:
			random.shuffle( window )
		
		metrics = []
		for d in [ dotlist[i] for i in window ]:
			rr = (d[0]-10+0.5) ** 2 + (d[1]-8+0.5) ** 2
			#metrics.append( -abs( 7**2 - rr ) )
			metrics.append( -rr )
		if bias > 0:
			return window[ metrics.index(max(metrics)) ]
		else:
			return window[ metrics.index(min(metrics)) ]


##############################################################################

class MultiInterval:
	def __init__ ( self ):
		self.ticks = 0
		self.ticklimit = 0
		self.intervals = set()
	
	def tick ( self ):
		self.ticks += 1
		if self.ticklimit > 0:
			self.ticks %= self.ticklimit
	
	def hit ( self, interval = 1 ):
		"""return true every interval'th tick"""
		if interval < 1:
			return False
		if interval == 1:
			return True
		if interval not in self.intervals:
			self.intervals.add( interval )
			self.ticklimit = 1
			for x in sorted( list(self.intervals), reverse = True ):
				if self.ticklimit % x != 0:
					self.ticklimit *= x
			# TODO more better ticklimit with prime factors hurble durble
		return self.ticks % interval == 0


def format_bytes ( bytes ):
#	if bytes < 100:
#		return '{0}'.format( int(bytes) )
	kb = bytes / float( 2**10 )
#	if kb < 100:
#		return '{0:.1f}K'.format( kb )
	if kb < 200:
		return '{0}K'.format( int(kb) )
	mb = kb / float( 2**10 )
	if mb < 10:
		return '{0:.2f}M'.format( mb )
	if mb < 1000:
		return '{0}M'.format( int( mb ) )
	gb = mb / float( 2**10 )
	return '{0}G'.format( int(gb) )


def format_duration ( seconds ):
	days = 0
	hours = 0
	minutes = 0
	if seconds >= 60 * 60 * 24:
		days = seconds / ( 60 * 60 * 24 )
		seconds -= days * 60 * 60 * 24
	if seconds >= 60 * 60:
		hours = seconds / ( 60 * 60 )
		seconds -= hours * 60 * 60
	if seconds >= 60:
		minutes = seconds / 60
		seconds -= minutes * 60
	#durstr = '{0}d{1}h{2}m{3}s'.format( days, hours, minutes, seconds )
	durstr = '{0}d {1}h {2}m'.format( days, hours, minutes )
	return durstr


def runpanel ():
	
	panel = PanelLink()
	stats = SystemStats()
	anim = LoadLevels()
	
	#                           01234567890123456789
	panel.LCD_putstring( 0, 0, '                    ' )
	panel.LCD_putstring( 1, 0, '                    ' )
	panel.LCD_putstring( 2, 0, '                    ' )
	panel.LCD_putstring( 3, 0, '                    ' )
	anim.updatepanel( panel, True )
	panel.sync()
	
	cpuload = 0
	diskload = None
	sched = MultiInterval()
	while True:
		
		if sched.hit( 20 ):
			ramstr = 'RAM {0}%  SW {1}%'.format( *stats.mem_load() )
			panel.LCD_putstring( 0, 0, '{0: <16}'.format( ramstr )[:16] )
		
		if sched.hit( 6 ):
			diskload = stats.disk_load()
		disks = ''.join([ '\xff' if sched.hit( 2 ) and diskload[dev] else dev[-1] for dev in sorted(diskload.keys()) ])
		panel.LCD_putstring( 1, 0, '{0: <16}'.format( 'DISK ' + disks )[:16] )
		
		if sched.hit( 50 ):
			netload = stats.net_load()
			netstr = 'NET {0}/{1}'.format( format_bytes(netload[1]*8), format_bytes(netload[0]*8) )
			panel.LCD_putstring( 2, 0, '{0: <16}'.format( netstr[:16] ) )
		
		if sched.hit( 10 ):
			cpuload = stats.CPU_load()
			#panel.LCD_putstring( 3, 10, '{0:.1f}%'.format( cpuload * 100 ) )
		anim.step( cpuload )
		anim.updatepanel( panel )
		
		if sched.hit( 50 ):
			panel.LCD_putstring( 2, 16, '{0:3}\xdf'.format( stats.CPU_temp() ) )
		
		if sched.hit( 50 ):
			upstr = format_duration( stats.uptime() )
			panel.LCD_putstring( 3, 0, '{0: <20}'.format( 'UP ' + upstr[:20] ) )
		
		#if sched.hit( 50 ):
		#	freestr = ''
		#	freestr += '/:' + format_bytes( stats.fs_free( '/' ) )
		#	freestr += ' ' + 'scratch:' + format_bytes( stats.fs_free( '/scratch' ) )
		#	panel.LCD_putstring( 3, 0, '{0: <20}'.format( freestr ) )
		
		sched.tick()
		time.sleep( 0.05 )


##############################################################################

def forkoff ():
	"""
	do the UNIX double-fork magic, see Stevens' "Advanced 
	Programming in the UNIX Environment" for details (ISBN 0201563177)
	http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
	"""
	try:
		pid = os.fork()
		if pid > 0:
			# exit first parent
			sys.exit(0)
	except OSError, e:
		sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
		sys.exit(1)
	
	# decouple from parent environment
	os.chdir("/")
	os.setsid()
	os.umask(0)
	
	# do second fork
	try:
		pid = os.fork()
		if pid > 0:
			# exit from second parent
			sys.exit(0)
	except OSError, e: 
		sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
		sys.exit(1)

	# redirect standard file descriptors
	sys.stdout.flush()
	sys.stderr.flush()
	si = file('/dev/null', 'r')
	so = file('/dev/null', 'a+')
	se = file('/dev/null', 'a+', 0)
	os.dup2(si.fileno(), sys.stdin.fileno())
	os.dup2(so.fileno(), sys.stdout.fileno())
	os.dup2(se.fileno(), sys.stderr.fileno())


def main ():
	if len(sys.argv) == 2 and sys.argv[1] == 'daemon':
		print 'forking to background'
		forkoff()
	
	runpanel()
	return 0

if __name__ == '__main__': main()
